@@ -1,11 +1,12 @@
 use crate::estimate_transaction_size::estimate_virtual_size;
 use crate::input::Input;
+use crate::oracle;
 use anyhow::{anyhow, bail, Context, Result};
-use bitcoin_hashes::hex::ToHex;
 use conquer_once::Lazy;
 use elements::bitcoin::{Amount, Network, PrivateKey, PublicKey};
 use elements::confidential::{self, Asset, AssetBlindingFactor, ValueBlindingFactor};
 use elements::encode::serialize;
+use elements::hashes::hex::ToHex;
 use elements::pset::serialize::Serialize;
 use elements::script::Builder;
 use elements::secp256k1_zkp::{self, Secp256k1, SecretKey, Signing, Verification, SECP256K1};
@@ -24,9 +25,6 @@ use std::future::Future;
 use std::str::FromStr;
 use std::time::SystemTime;
 
-#[cfg(test)]
-mod protocol_tests;
-
 /// Secret key used to produce a signature which proves that an
 /// input's witness stack contains transaction data equivalent to the
 /// transaction which includes the input itself.
@@ -87,6 +85,7 @@ pub struct CollateralContract {
     repayment_principal_output: TxOut,
     repayment_principal_output_blinder: SecretKey,
     timelock: u32,
+    chain: Chain,
     // Dynamic liquidation
     oracle_pk: PublicKey,
     min_price_btc: u64,
@@ -104,6 +103,7 @@ impl CollateralContract {
     const COV_SCRIPT_BYTES: [u8; 3] = [0xad, 0xc1, 0x68];
 
     /// Fill in the collateral contract template with the provided arguments.
+    #[allow(clippy::too_many_arguments)]
     fn new(
         borrower_pk: PublicKey,
         lender_pk: PublicKey,
@@ -112,6 +112,7 @@ impl CollateralContract {
         oracle_pk: PublicKey,
         min_price_btc: u64,
         min_timestamp: u64,
+        chain: Chain,
     ) -> Result<Self> {
         use elements::opcodes::all::*;
 
@@ -195,6 +196,7 @@ impl CollateralContract {
             repayment_principal_output,
             repayment_principal_output_blinder,
             timelock,
+            chain,
             oracle_pk,
             min_price_btc,
             min_timestamp,
@@ -209,7 +211,7 @@ impl CollateralContract {
         identity_signer: S,
     ) -> Result<()>
     where
-        S: FnOnce(secp256k1::Message) -> SF,
+        S: FnOnce(secp256k1_zkp::Message) -> SF,
         SF: Future<Output = Result<Signature>>,
     {
         let transaction_cloned = transaction.clone();
@@ -241,7 +243,7 @@ impl CollateralContract {
         identity_signer: S,
     ) -> Result<()>
     where
-        S: FnOnce(secp256k1::Message) -> SF,
+        S: FnOnce(secp256k1_zkp::Message) -> SF,
         SF: Future<Output = Result<Signature>>,
     {
         let transaction_cloned = transaction.clone();
@@ -276,7 +278,7 @@ impl CollateralContract {
         oracle_sig: Signature,
     ) -> Result<()>
     where
-        S: FnOnce(secp256k1::Message) -> SF,
+        S: FnOnce(secp256k1_zkp::Message) -> SF,
         SF: Future<Output = Result<Signature>>,
     {
         let btc_price = oracle_msg.price_to_bytes();
@@ -328,7 +330,7 @@ impl CollateralContract {
         cov_script: &'a Script,
     ) -> Result<impl Satisfier<PublicKey> + 'a>
     where
-        S: FnOnce(secp256k1::Message) -> SF,
+        S: FnOnce(secp256k1_zkp::Message) -> SF,
         SF: Future<Output = Result<Signature>>,
     {
         let descriptor_cov = &self.descriptor.as_cov().expect("covenant descriptor");
@@ -402,13 +404,12 @@ impl CollateralContract {
         Ok(())
     }
 
-    fn address(&self) -> Address {
-        // FIXME: AddressParams should not be hard-coded
-        Address::p2wsh(&self.raw_script, None, &AddressParams::ELEMENTS)
+    pub fn address(&self) -> Address {
+        Address::p2wsh(&self.raw_script, None, self.chain.into())
     }
 
     fn blinded_address(&self, blinder: secp256k1_zkp::PublicKey) -> Address {
-        Address::p2wsh(&self.raw_script, Some(blinder), &AddressParams::ELEMENTS)
+        Address::p2wsh(&self.raw_script, Some(blinder), self.chain.into())
     }
 
     fn repayment_amount<C>(&self, secp: &Secp256k1<C>) -> Result<Amount>
@@ -436,12 +437,12 @@ pub struct LoanRequest {
     #[serde(with = "::elements::bitcoin::util::amount::serde::as_sat")]
     fee_sats_per_vbyte: Amount,
     borrower_pk: PublicKey,
-    timelock: u32,
     borrower_address: Address,
 }
 
 impl LoanRequest {
     /// Get a copy of the collateral amount.
+    #[deprecated(note = "Use accessor on Borrower0 instead", since = "0.3.0")]
     pub fn collateral_amount(&self) -> Amount {
         self.collateral_amount
     }
@@ -475,7 +476,6 @@ pub struct Borrower0 {
     collateral_inputs: Vec<Input>,
     #[serde(with = "::elements::bitcoin::util::amount::serde::as_sat")]
     fee_sats_per_vbyte: Amount,
-    timelock: u32,
     bitcoin_asset_id: AssetId,
     usdt_asset_id: AssetId,
 }
@@ -489,7 +489,6 @@ impl Borrower0 {
         address_blinding_sk: SecretKey,
         collateral_amount: Amount,
         fee_sats_per_vbyte: Amount,
-        timelock: u32,
         bitcoin_asset_id: AssetId,
         usdt_asset_id: AssetId,
     ) -> Result<Self>
@@ -508,19 +507,18 @@ impl Borrower0 {
             collateral_amount,
             collateral_inputs,
             fee_sats_per_vbyte,
-            timelock,
             bitcoin_asset_id,
             usdt_asset_id,
         })
     }
 
+    #[deprecated(note = "Use accessors on Self instead", since = "0.3.0")]
     pub fn loan_request(&self) -> LoanRequest {
         LoanRequest {
             collateral_amount: self.collateral_amount,
             collateral_inputs: self.collateral_inputs.clone(),
             fee_sats_per_vbyte: self.fee_sats_per_vbyte,
             borrower_pk: self.keypair.1,
-            timelock: self.timelock,
             borrower_address: self.address.clone(),
         }
     }
@@ -619,6 +617,34 @@ impl Borrower0 {
             usdt_asset_id: self.usdt_asset_id,
         })
     }
+
+    /// Get a reference to the value of the collateral the borrower
+    /// will put up for the loan.
+    pub fn collateral_amount(&self) -> &Amount {
+        &self.collateral_amount
+    }
+
+    /// Get a reference to the inputs the borrower will use to fund
+    /// the loan transaction's collateral.
+    pub fn collateral_inputs(&self) -> &[Input] {
+        self.collateral_inputs.as_slice()
+    }
+
+    /// Get a copy of how many satoshis per vbyte the borrower will
+    /// use to pay for the loan transaction's fees.
+    pub fn fee_sats_per_vbyte(&self) -> Amount {
+        self.fee_sats_per_vbyte
+    }
+
+    /// Get a reference to the borrower's address.
+    pub fn address(&self) -> &Address {
+        &self.address
+    }
+
+    /// Get a copy of the borrower's PublicKey.
+    pub fn pk(&self) -> PublicKey {
+        self.keypair.1
+    }
 }
 
 #[derive(Debug, Clone, serde::Serialize, serde::Deserialize)]
@@ -744,6 +770,8 @@ impl Borrower1 {
             .map(|input| input.txin)
             .collect();
         tx_ins.push(collateral_input.txin);
+        let collateral_contract_index = (tx_ins.len() - 1) as u32;
+
         let tx_ins = tx_ins
             .into_iter()
             .map(|previous_output| TxIn {
@@ -816,7 +844,7 @@ impl Borrower1 {
             .satisfy_loan_repayment(
                 &mut tx,
                 self.repayment_collateral_input.original_txout.value,
-                1,
+                collateral_contract_index,
                 // TODO: Push ownership (and generation) of secret key outside of the library
                 |message| async move { Ok(SECP256K1.sign(&message, &self.keypair.0)) },
             )
@@ -886,12 +914,14 @@ impl Lender0 {
     /// Interpret a loan request and performs lender logic.
     ///
     /// rate is expressed in usdt sats per btc, i.e. rate = 1 BTC / USDT
+    #[deprecated(note = "Use interpret_loan_request instead", since = "0.3.0")]
     pub async fn interpret<R, C, CS, CF>(
         self,
         rng: &mut R,
         secp: &Secp256k1<C>,
         coin_selector: CS,
         loan_request: LoanRequest,
+        timelock: u32,
         rate: u64,
     ) -> Result<Lender1>
     where
@@ -900,14 +930,51 @@ impl Lender0 {
         CS: FnOnce(Amount, AssetId) -> CF,
         CF: Future<Output = Result<Vec<Input>>>,
     {
-        // TODO: This API should change so that the caller can decide on all this outside of this library
         let LoanAmounts {
             principal: principal_amount,
             repayment: repayment_amount,
-            min_price_btc,
-        } = Lender0::calculate_loan_amounts(&loan_request, rate, 20, 10)?;
-        let collateral_inputs = loan_request
-            .collateral_inputs
+            min_collateral_price,
+        } = Lender0::calculate_loan_amounts(loan_request.collateral_amount, rate, 20, 10)?;
+
+        let principal_inputs = coin_selector(principal_amount, self.usdt_asset_id).await?;
+
+        self.build_loan_transaction(
+            rng,
+            secp,
+            loan_request.fee_sats_per_vbyte,
+            (
+                loan_request.collateral_amount,
+                loan_request.collateral_inputs,
+            ),
+            (principal_amount, principal_inputs),
+            repayment_amount,
+            min_collateral_price,
+            (loan_request.borrower_pk, loan_request.borrower_address),
+            timelock,
+        )
+        .await
+    }
+
+    #[allow(clippy::too_many_arguments)]
+    pub async fn build_loan_transaction<R, C>(
+        self,
+        rng: &mut R,
+        secp: &Secp256k1<C>,
+        fee_sats_per_vbyte: Amount,
+        (collateral_amount, collateral_inputs): (Amount, Vec<Input>),
+        (principal_amount, principal_inputs): (Amount, Vec<Input>),
+        repayment_amount: Amount,
+        min_collateral_price: u64,
+        (borrower_pk, borrower_address): (PublicKey, Address),
+        timelock: u32,
+    ) -> Result<Lender1>
+    where
+        R: RngCore + CryptoRng,
+        C: Verification + Signing,
+    {
+        let chain = Chain::new(&borrower_address, &self.address)?;
+
+        let collateral_inputs = collateral_inputs
             .into_iter()
             .map(|input| input.into_unblinded_input(secp))
             .collect::<Result<Vec<_>>>()?;
@@ -916,7 +983,6 @@ impl Lender0 {
             .iter()
             .map(|input| (input.txout.asset, &input.secrets));
 
-        let principal_inputs = coin_selector(principal_amount, self.usdt_asset_id).await?;
         let unblinded_principal_inputs = principal_inputs
             .clone()
             .into_iter()
@@ -933,8 +999,6 @@ impl Lender0 {
             .iter()
             .fold(0, |sum, input| sum + input.secrets.value);
 
-        let collateral_amount = loan_request.collateral_amount;
-
         let (repayment_principal_output, repayment_collateral_abf, repayment_collateral_vbf) = {
             let dummy_asset_id = self.usdt_asset_id;
             let dummy_abf = AssetBlindingFactor::new(rng);
@@ -965,13 +1029,14 @@ impl Lender0 {
         let contract_creation_timestamp = now.as_secs() + 300;
 
         let collateral_contract = CollateralContract::new(
-            loan_request.borrower_pk,
+            borrower_pk,
             lender_pk,
-            loan_request.timelock,
+            timelock,
             (repayment_principal_output, self.address_blinder),
             self.oracle_pk,
-            min_price_btc,
+            min_collateral_price,
             contract_creation_timestamp,
+            chain,
         )
         .context("could not build collateral contract")?;
         let collateral_address = collateral_contract.blinded_address(collateral_blinding_pk.key);
@@ -994,7 +1059,7 @@ impl Lender0 {
             rng,
             secp,
             principal_amount.as_sat(),
-            loan_request.borrower_address.clone(),
+            borrower_address.clone(),
             self.usdt_asset_id,
             inputs_not_last_confidential.as_slice(),
         )
@@ -1037,8 +1102,7 @@ impl Lender0 {
         ];
 
         let tx_fee = Amount::from_sat(
-            estimate_virtual_size(inputs.len() as u64, 4)
-                * loan_request.fee_sats_per_vbyte.as_sat(),
+            estimate_virtual_size(inputs.len() as u64, 4) * fee_sats_per_vbyte.as_sat(),
         );
         let collateral_change_amount = Amount::from_sat(collateral_input_amount)
             .checked_sub(collateral_amount)
@@ -1054,7 +1118,7 @@ impl Lender0 {
             rng,
             secp,
             collateral_change_amount.as_sat(),
-            loan_request.borrower_address,
+            borrower_address,
             self.bitcoin_asset_id,
             inputs
                 .iter()
@@ -1117,10 +1181,10 @@ impl Lender0 {
         Ok(Lender1 {
             keypair: self.keypair,
             address: self.address,
-            timelock: loan_request.timelock,
+            timelock,
             loan_transaction,
             collateral_contract,
-            collateral_amount: loan_request.collateral_amount,
+            collateral_amount,
             repayment_collateral_input,
             repayment_collateral_abf,
             repayment_collateral_vbf,
@@ -1129,7 +1193,7 @@ impl Lender0 {
     }
 
     fn calculate_loan_amounts(
-        loan_request: &LoanRequest,
+        collateral_amount: Amount,
         rate: u64,
         overcollateralization_percentage: u8,
         interest_percentage: u8,
@@ -1137,7 +1201,7 @@ impl Lender0 {
         use rust_decimal::prelude::ToPrimitive;
         use rust_decimal::Decimal;
 
-        let sats = loan_request.collateral_amount.as_sat();
+        let sats = collateral_amount.as_sat();
         let btc = Decimal::from(sats)
             .checked_div(Decimal::from(Amount::ONE_BTC.as_sat()))
             .context("division overflow")?;
@@ -1163,7 +1227,7 @@ impl Lender0 {
         let repayment = Decimal::from(principal.as_sat()) * interest_factor;
 
         let min_price_btc = repayment.checked_div(btc).context("division overflow")?;
-        let min_price_btc = min_price_btc
+        let min_collateral_price = min_price_btc
             .to_u64()
             .context("decimal cannot be represented as u64")?;
 
@@ -1175,7 +1239,7 @@ impl Lender0 {
         Ok(LoanAmounts {
             principal,
             repayment,
-            min_price_btc,
+            min_collateral_price,
         })
     }
 }
@@ -1187,7 +1251,7 @@ struct LoanAmounts {
     repayment: Amount,
     /// Price under which the lender will be able to unilaterally
     /// liquidate the collateral contract, in whole USD.
-    min_price_btc: u64,
+    min_collateral_price: u64,
 }
 
 pub struct Lender1 {
@@ -1373,70 +1437,6 @@ impl Lender1 {
     }
 }
 
-/// Toy version of the oracle module used to define the parts of the
-/// oracle which matter to the loan protocol.
-///
-/// In particular, the oracle must encode the message in such a way
-/// that it can be decomposed and used from within an Elements script.
-pub(crate) mod oracle {
-    pub struct Message {
-        /// Price of bitcoin in whole USD.
-        btc_price: WitnessStackInteger,
-        /// UNIX timestamp.
-        timestamp: WitnessStackInteger,
-    }
-
-    impl Message {
-        pub fn new(btc_price: u64, timestamp: u64) -> Self {
-            Self {
-                btc_price: WitnessStackInteger(btc_price),
-                timestamp: WitnessStackInteger(timestamp),
-            }
-        }
-
-        /// Serialize price as bytes.
-        pub fn price_to_bytes(&self) -> Vec<u8> {
-            self.btc_price.serialize()
-        }
-
-        /// Serialize timestamp as bytes.
-        pub fn timestamp_to_bytes(&self) -> Vec<u8> {
-            self.timestamp.serialize()
-        }
-
-        #[cfg(test)]
-        pub fn message_hash(&self) -> secp256k1::Message {
-            use bitcoin_hashes::{sha256, Hash, HashEngine};
-
-            let mut sha256d = sha256::Hash::engine();
-            sha256d.input(&self.price_to_bytes());
-            sha256d.input(&self.timestamp_to_bytes());
-            let message_hash = sha256::Hash::from_engine(sha256d);
-
-            secp256k1::Message::from_slice(&message_hash).unwrap()
-        }
-    }
-
-    struct WitnessStackInteger(u64);
-
-    impl WitnessStackInteger {
-        /// Serialize an integer so that it can be included in a Bitcoin witness stack.
-        ///
-        /// Said format is a little-endian byte encoding without trailing 0-bytes.
-        fn serialize(&self) -> Vec<u8> {
-            // to save a reverse operation, we first encode it as big-endian
-            let bytes = self.0.to_be_bytes().to_vec();
-            let mut bytes = bytes
-                .into_iter()
-                .skip_while(|byte| *byte == 0)
-                .collect::<Vec<_>>();
-            bytes.reverse();
-
-            bytes
-        }
-    }
-}
-
 fn make_keypair<R>(rng: &mut R) -> (SecretKey, PublicKey)
 where
     R: RngCore + CryptoRng,
@@ -1473,6 +1473,57 @@ pub mod transaction_as_string {
     }
 }
 
+/// Possible networks on which the loan contract may be deployed.
+#[derive(Debug, Clone, Copy, serde::Serialize, serde::Deserialize)]
+enum Chain {
+    Elements,
+    Liquid,
+}
+
+impl Chain {
+    /// Calculate the `Chain` based on the addresses of the borrower and the lender.
+    ///
+    /// If their `AddressParams` don't match it means that the parties
+    /// are not on the same network, so we return an error.
+    ///
+    /// If their `AddressParams` don't describe a known chain, we
+    /// return an error.
+    fn new(borrower: &Address, lender: &Address) -> Result<Self, ChainError> {
+        let borrower = borrower.params;
+        let lender = lender.params;
+
+        if borrower != lender {
+            return Err(ChainError::Different { borrower, lender });
+        }
+
+        match *borrower {
+            AddressParams::ELEMENTS => Ok(Self::Elements),
+            AddressParams::LIQUID => Ok(Self::Liquid),
+            _ => Err(ChainError::Custom(borrower)),
+        }
+    }
+}
+
+impl From<Chain> for &AddressParams {
+    fn from(from: Chain) -> Self {
+        match from {
+            Chain::Elements => &AddressParams::ELEMENTS,
+            Chain::Liquid => &AddressParams::LIQUID,
+        }
+    }
+}
+
+#[derive(thiserror::Error, Debug)]
+enum ChainError {
+    #[error("Borrower and lender are on different chains: {borrower:?} vs {lender:?}")]
+    Different {
+        borrower: &'static AddressParams,
+        lender: &'static AddressParams,
+    },
+    #[error("Unsupported custom chain: {0:?}")]
+    Custom(&'static AddressParams),
+}
+
 #[cfg(test)]
 mod constant_tests {
     use super::*;
